from demessaging.PulsarMessageConsumer import PulsarMessageConsumer
from demessaging.PulsarMessageConstants import PulsarConfigKeys
from demessaging.progress_send_handler import ProgressSendHandler
from digitalearth_floodmask_update.update_flood_events_routine import FLOOD_UPDATE_PIPELINE
from deprogressapi import ProgressReport
import json
import sys
import asyncio


class CifiMessageHandler:
    __pulsarConfig = {
        PulsarConfigKeys.HOST: 'rz-vm154.gfz-potsdam.de',
        PulsarConfigKeys.PORT: '8082',
        PulsarConfigKeys.PERSISTENT: 'non-persistent',
        PulsarConfigKeys.TENANT: 'public',
        PulsarConfigKeys.NAMESPACE: 'digital-earth',
        PulsarConfigKeys.TOPIC: 'cifi',
    }

    __cifi_pipeline_config = {
        "base_url": "https://emergency.copernicus.eu",
        "activations_list_url": "https://emergency.copernicus.eu/mapping/list-of-activations-rapid",
        "activation_filters": {"affected_countries": ["DE"]},
        "geoserver_url": "http://rz-vm154.gfz-potsdam.de:8080/geoserver",
        "username": "***enter-geoserver-username-here***",
        "password": "***enter-geoserver-password-here***",
        "workspace": "flood",
        "digital_elevation_model": "digitalearth:de_dem25m",
        "enforce_xy_order": True,
        "min_recency_in_days": 30
    }

    def __init__(self):
        self.__pulsar = PulsarMessageConsumer(pulsar_config=self.__pulsarConfig, handle_request=self.handle_message)

    def start(self):
        self.__pulsar.connect()
        self.__pulsar.wait_for_request()

    def stop(self):
        self.__pulsar.disconnect()

    def handle_message(self, request_msg):
        try:
            print('handling message in message handler')
            # extract request content, e.g. from message properties or message payload
            props = request_msg['properties']

            if 'function' not in props:
                raise Exception('missing function name in request properties')

            requested_function = props['function']
            requested_function_args = json.loads(props['function_args'])

            # here we distinguish what to do based on the requested function, we have implemented several approaches
            # for this, but since we only have one function yet, we keep things simple and call it directly
            if requested_function == 'update_database':
                # initialize the progress send handler and the root progress report
                # howto produce progress updates see: https://pypi.org/project/deprogressapi/

                if 'min_recency_in_days' not in requested_function_args:
                    raise AttributeError("missing function argument: 'min_recency_in_days'")

                # extract the num days argument and store it in the config
                cfg = self.__cifi_pipeline_config.copy()
                cfg['min_recency_in_days'] = int(requested_function_args['min_recency_in_days'])

                send_handler = ProgressSendHandler(pulsar=self.__pulsar, request_msg=request_msg)
                report = ProgressReport(step_message='update database', send_handler=send_handler,
                                        steps=len(FLOOD_UPDATE_PIPELINE.stages) + 1)

                try:
                    FLOOD_UPDATE_PIPELINE.set_progress_report(report)

                    # workaround to fix issues with running requests-html outside main thread
                    async def run_pipeline():
                        return FLOOD_UPDATE_PIPELINE.initialize_and_run(cfg)

                    loop = asyncio.new_event_loop()
                    asyncio.set_event_loop(loop)
                    loop.run_until_complete(run_pipeline())

                    # send response
                    response_payload = None
                    self.__pulsar.send_response(request=request_msg,
                                                response_properties={'status': 'success'},
                                                response_payload=response_payload)
                    report.step_message = "Database update complete."
                    report.complete()

                except Exception as e:
                    error = str(e)
                    report.step_message = f"error '{report.step_message}': {error}"
                    report.complete()
                    self.__pulsar.send_error(request=request_msg, error_message=error)
                    print(error)
        except Exception as e:
            print(e, file=sys.stderr)
            self.send_error(request_msg, str(e))
            return

    def send_error(self, request, error_message):
        self.__pulsar.send_response(request=request,
                                    response_payload=error_message,
                                    response_properties={'status': 'error'})
        print(error_message)
