import requests
from geo.Geoserver import Geoserver
import tempfile
from pathlib import Path
import os
import zipfile


class FloodGeoserver(Geoserver):
    """
    Extends the Geoserver class of the geo package by some utility functions
    """

    def create_vector_datastore_from_in_memory_zip(self, in_memory_zip, store_name, workspace=None,
                                                   file_extension="shp"):
        """
        Create datastore from an in-memory zipfile containing a shapefile and related files.
        Parameters
        ----------
        in_memory_zip : BytesIO object of the zipped shapefile
        store_name : str
            Name of store to be created
        workspace: str, optional
            Name of workspace to be used. Default: "default".
        file_extension : str
        Notes
        -----
        The layer name will be assigned according to the shp name
        """
        try:
            if workspace is None:
                workspace = "default"

            headers = {
                "Content-type": "application/zip",
                "Accept": "application/xml",
            }

            url = "{0}/rest/workspaces/{1}/datastores/{2}/file.{3}?filename={2}&update=overwrite".format(
                self.service_url, workspace, store_name, file_extension
            )

            data = in_memory_zip.getvalue()
            r = requests.put(
                url,
                data=data,
                auth=(self.username, self.password),
                headers=headers,
            )

            if r.status_code in [200, 201]:
                return "The shapefile datastore created successfully!"

            else:
                # with open(f"{store_name}_error_status_code_{r.status_code}.zip", "wb") as f:
                #    f.write(in_memory_zip.getbuffer())
                return "{}: The shapefile datastore can not be created!".format(
                    r.status_code
                )

        except Exception as e:
            return "Error: {}".format(e)

    def create_shp_datastore_from_geodataframe(self, geodataframe, shp_name, workspace):
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_dir = Path(temp_dir)
            geodataframe.to_file(str(temp_dir / shp_name))
            shapefiles = os.listdir(temp_dir)
            zip_name = shp_name.replace(".shp", ".zip")
            zip_path = str(temp_dir / zip_name)
            with zipfile.ZipFile(zip_path, 'w') as new_archive:
                for shapefile in shapefiles:
                    p = str(temp_dir / shapefile)
                    with open(p, "rb") as f:
                        file_bytes = f.read()
                        new_archive.writestr(shapefile, file_bytes, compress_type=zipfile.ZIP_DEFLATED)
            msg = self.create_shp_datastore(path=zip_path, workspace=workspace)
            return msg

    def update_datastore_description(self, workspace, store_name, description):
        """Updates description of target datastore"""
        url = "{0}/rest/workspaces/{1}/datastores/{2}.json".format(self.service_url, workspace, store_name)
        auth = (self.username, self.password)
        new_description = {'dataStore': {'description': description}}
        resp = requests.put(url=url, json=new_description, auth=auth)
        return resp.status_code

    def update_featuretype_description(self, workspace, store_name, featuretype_name, description):
        """
        Updates description of target featuretype
        """
        url = "{}/rest/workspaces/{}/datastores/{}/featuretypes/{}.json".format(
            self.service_url, workspace, store_name, featuretype_name
        )
        new_description = {"featureType": {"abstract": description}}
        r = requests.put(url=url, json=new_description, auth=(self.username, self.password))
        return r.status_code
