import asyncio
from random import randint
from datetime import datetime, timedelta, date
import time
from requests.exceptions import RequestException
from urllib.parse import urljoin
from requests_html import HTMLSession, AsyncHTMLSession
import pandas as pd
import nest_asyncio


class CopernicusScraper:
    """
    Utility class to scrape a list of available activations, activation components and download products from the
    Copernicus Emergency Management Service (EMS)
    """

    EVENT_TYPE_TID_MAPPING = {
        'Drought': 617,
        'Epidemic': 618,
        'Extreme temperature': 616,
        'Humanitarian': 520,
        'Infestation': 619,
        'Mass movement': 580,
        'Storm': 597,
        'Transportation accident': 620,
        'Volcanic activity': 615,
        'Wildfire': 535,
        'Forest fire, wild fire': 16,
        'Flood': 17,
        'Windstorm': 18,
        'Earthquake': 19,
        'Industrial accident': 20,
        'Other': 21
    }

    def __init__(self, base_url, activations_list_url, activations=None, random_wait_time_min_sec=1,
                 random_wait_time_max_sec=3, progress_report=None, session=None):
        self.base_url = base_url
        self.activations_list_url = activations_list_url
        if session is None:
            self.session = HTMLSession()
        else:
            self.session = session
        self.activations = activations
        if random_wait_time_min_sec > random_wait_time_max_sec:
            random_wait_time_min_sec, random_wait_time_max_sec = random_wait_time_max_sec, random_wait_time_min_sec
        self.random_wait_time_min_sec = random_wait_time_min_sec
        self.random_wait_time_max_sec = random_wait_time_max_sec
        self.progress_report = progress_report

    def get_all_forms(self, url, render_js=True):
        """Returns all form tags found on a web page's url """
        res = self.__send_request(url, render_js=render_js)
        return res.html.find('form')

    @classmethod
    def get_form_details(cls, form):
        """Returns the HTML details of a form, including action, method and list of form controls (inputs, etc)"""
        details = {}
        # get the form action (requested URL)
        action = form.attrs.get('action').lower()
        # get the form method (POST, GET, DELETE, etc)
        # if not specified, GET is the default in HTML
        method = form.attrs.get('method', 'get').lower()
        # get all form inputs
        inputs = []
        for input_tag in form.find('input'):
            # get type of input form control
            input_type = input_tag.attrs.get('type', 'text')
            # get name attribute
            input_name = input_tag.attrs.get('name')
            # get the default value of that input tag
            input_value = input_tag.attrs.get('value', '')
            # add everything to that list
            inputs.append({'type': input_type, 'name': input_name, 'value': input_value})
        # put everything to the resulting dictionary
        details['action'] = action
        details['method'] = method
        details['inputs'] = inputs
        return details

    def request_product_data(self, product_download_url, method='post', render_js=True):
        """Downloads product data from the given url"""
        # product downloads are hidden behind a no liability disclaimer...
        # prepare filling out the form
        disclaimer_form = self.get_all_forms(product_download_url, render_js=render_js)[-1]
        disclaimer_form_details = self.get_form_details(disclaimer_form)
        disclaimer_form_data = {}
        for input_tag in disclaimer_form_details['inputs']:
            if input_tag['type'] != 'submit':
                disclaimer_form_data[input_tag['name']] = input_tag['value']

        # send download request with filled out form
        if method == 'post':
            if isinstance(self.session, AsyncHTMLSession):
                current_loop = asyncio.get_running_loop()
                if current_loop.is_running():
                    nest_asyncio.apply(current_loop)
                res = current_loop.run_until_complete(
                    self.session.post(product_download_url, data=disclaimer_form_data, stream=True)
                )
            else:
                res = self.session.post(product_download_url, data=disclaimer_form_data, stream=True)
        else:
            if isinstance(self.session, AsyncHTMLSession):
                current_loop = asyncio.get_running_loop()
                if current_loop.is_running():
                    nest_asyncio.apply(current_loop)
                res = current_loop.run_until_complete(
                    self.session.get(product_download_url, data=disclaimer_form_data, stream=True)
                )
            else:
                res = self.session.get(product_download_url, params=disclaimer_form_data, stream=True)
        return res

    @classmethod
    def __build_url_filter_tail(cls, title_op, title, field_event_type_tid,
                                event_time_utc_start, event_time_utc_end, activation_status,
                                affected_countries):
        """Build parameter string of url to filter table results of rapid mapping activations"""

        parameter_template = \
            '?title_op={}&title={}&field_event_type_tid%5B%5D={}&field_event_time_utc_value%5Bmin%5D%5Bdate%5D={}&' + \
            'field_event_time_utc_value%5Bmax%5D%5Bdate%5D={}&field_activation_status_value={}'

        if not event_time_utc_end:
            next_day = date.today() + timedelta(days=1)
            event_time_utc_end = datetime.strftime(next_day, '%Y-%m-%d')

        if affected_countries:
            ac_param_string = ''
            for affected_country in affected_countries:
                ac_param_string += ('&affected_countries_list%5B%5D=' + affected_country)
            parameter_template += ac_param_string

        parameterized_request_tail = parameter_template.format(title_op, title, field_event_type_tid,
                                                               event_time_utc_start, event_time_utc_end,
                                                               activation_status)
        return parameterized_request_tail

    def __send_request(self, url, render_js=True):
        """
        helper function to send a request to given url depending on whether an async or sync session is used

        this whole function was written to patch the issues with using requests-html outside the main thread in a
        multi-threaded environment
        """
        if isinstance(self.session, AsyncHTMLSession):
            # requests-html requires creating an event loop but asyncio doesn't allow nested loops
            # therefore patch asyncio loop if not executed in main thread
            current_loop = asyncio.get_running_loop()
            if current_loop.is_running():
                nest_asyncio.apply(current_loop)
            response = current_loop.run_until_complete(self.session.get(url))
            # run all js and update html object
            # this will download Chromium Driver (first time executed only)
            if render_js:
                current_loop.run_until_complete(response.html.arender())
        else:
            response = self.session.get(url)
            # run all js and update html object
            if render_js:
                response.html.render()  # this will download Chromium Driver (first time executed only)
        return response

    def scrape_activations_list_url(self, title_op='contains', title='', field_event_type_tid=17,
                                    event_time_utc_start='', event_time_utc_end='', activation_status='All',
                                    affected_countries=[]):
        try:
            url_filter_tail = self.__build_url_filter_tail(title_op, title, field_event_type_tid, event_time_utc_start,
                                                           event_time_utc_end, activation_status, affected_countries)
            activations_url_with_filters = urljoin(self.activations_list_url, url_filter_tail)
            response = self.__send_request(activations_url_with_filters)
        except RequestException as e:
            return False, str(e)
        except Exception as e:
            return False, str(e)

        try:
            html = response.html
            # get headers of table
            th_tags = html.find('th')
            table_headers = [th.text.strip() for th in th_tags][1:-1]  # don't need 1st and last column
            table_headers.append('Component List Link')
            # get table rows
            table_tag = html.find('table', first=True)
            if table_tag is None:
                raise Exception('no data found for given filter criteria')
            tbody_tag = table_tag.find('tbody', first=True)
            if tbody_tag is None:
                raise Exception('no data found for given filter criteria')
            tr_tags = tbody_tag.find('tr')  # tr tags mark table rows
            if tr_tags is None:
                raise Exception('no data found for given filter criteria')

            table_rows = []
            for tr in tr_tags:
                td_tags = tr.find('td')  # td tags mark table cells
                row = [td.text.strip() for td in td_tags][1:-1]  # don't need 1st and last column
                relative_url = td_tags[2].find('a', first=True).attrs['href']
                absolute_url = urljoin(self.base_url, relative_url)
                row.append(absolute_url)
                table_rows.append(row)

            activations_dataframe = pd.DataFrame(table_rows, columns=table_headers)
            self.activations = activations_dataframe
            return True, ""
        except Exception as e:
            return False, str(e)

    @classmethod
    def __search_by_waypoints(cls, root, first_level_waypoint, second_level_waypoint=None):
        """Two-step search for HTML elements using CSS-selectors"""
        try:
            if second_level_waypoint is None:
                elem = root.find(first_level_waypoint, first=True)
            else:
                elem = root.find(first_level_waypoint, first=True).find(second_level_waypoint, first=True)
            return elem, ""
        except AttributeError as e:
            return None, 'e, | failed tags: {0}, {1}'.format(first_level_waypoint, second_level_waypoint)

    def __scrape_meta_root_fields(self, activation_code, meta_root):
        """Scrapes all meta information from the top of the EMS Activations component page"""

        component_title, _ = self.__search_by_waypoints(meta_root, '.views-field-title', '.field-content')
        if component_title:
            component_title = component_title.text
            component_id = component_title.split(':')[0]
        else:
            component_title = activation_code
            component_id = activation_code

        component_event_time_utc, _ = self.__search_by_waypoints(meta_root, '.views-field-field-event-time-utc',
                                                                 '.date-display-single')
        if component_event_time_utc:
            component_event_time_utc = component_event_time_utc.attrs['content']

        component_event_time_loc, _ = self.__search_by_waypoints(meta_root, '.views-field-field-event-time-loc',
                                                                 '.date-display-single')
        if component_event_time_loc:
            component_event_time_loc = component_event_time_loc.attrs['content']

        component_event_type, _ = self.__search_by_waypoints(meta_root, '.views-field-field-event-type-full',
                                                             '.field-content')
        if component_event_type:
            component_event_type = component_event_type.text

        component_activation_time_utc, _ = self.__search_by_waypoints(meta_root, '.views-field-field-activation-time-utc',
                                                                      '.date-display-single')
        if component_activation_time_utc:
            component_activation_time_utc = component_activation_time_utc.attrs['content']

        component_activation_status, _ = self.__search_by_waypoints(meta_root, '.views-field-field-activation-status',
                                                                    '.field-content')
        if component_activation_status:
            component_activation_status = component_activation_status.text

        component_affected_countries, _ = self.__search_by_waypoints(meta_root, '.views-field-field-affected-countries',
                                                                     '.field-content')
        if component_affected_countries:
            component_affected_countries = component_affected_countries.text

        component_authorised_user, _ = self.__search_by_waypoints(meta_root, '.views-field-field-activator',
                                                                  '.field-content')
        if component_authorised_user:
            component_authorised_user = component_authorised_user.text

        component_activation_reason, _ = self.__search_by_waypoints(meta_root, '.views-field-field-activation-reason',
                                                                    '.field-content')
        if component_activation_reason:
            component_activation_reason = component_activation_reason.text

        component = {'component id': component_id,
                     'meta': {
                         'component title': component_title,
                         'event time (utc)': component_event_time_utc,
                         'event time (loc)': component_event_time_loc,
                         'event type': component_event_type,
                         'activation time (utc)': component_activation_time_utc,
                         'activation status': component_activation_status,
                         'affected countries': component_affected_countries,
                         'authorised user': component_authorised_user,
                         'activation reason': component_activation_reason}
                     }
        return component

    def __scrape_component_products(self, activation_code, products_root):
        """Scrape all available products for a given component page"""
        product_rows = products_root.find('.views-row')
        component_products = {}
        for idx, product_row in enumerate(product_rows):
            product_name, _ = self.__search_by_waypoints(product_row, '.views-field-title', 'a')
            if product_name:
                product_name = product_name.text
                if "waiting for data" in product_name:
                    # this are just placeholders for products which do not yet have any data in them
                    continue
            else:
                product_name = "{}_product_{}".format(activation_code, idx)

            product_delivery_time, _ = self.__search_by_waypoints(product_row,
                                                                  '.views-field-field-component-delivery-time',
                                                                  '.field-content')
            if product_delivery_time:
                product_delivery_time = product_delivery_time.text

            product_version, _ = self.__search_by_waypoints(product_row, '.views-field-nothing-13', '.field-content')
            if product_version:
                product_version = product_version.text.split(':')[-1].strip()

            product_status, _ = self.__search_by_waypoints(product_row, '.views-field-nothing-14', '.field-content')
            if product_status:
                product_status = product_status.text.split(':')[-1].strip()

            product_download_pdf, _ = self.__search_by_waypoints(product_row, 'div[class$="pdf"]', 'a')
            if product_download_pdf:
                product_download_pdf = list(product_download_pdf.absolute_links)[0]

            product_download_jpeg, _ = self.__search_by_waypoints(product_row, 'div[class$="jpeg"]', 'a')
            if product_download_jpeg:
                product_download_jpeg = list(product_download_jpeg.absolute_links)[0]

            product_download_vector, _ = self.__search_by_waypoints(product_row,
                                                                    '.views-field-field-component-file-vectors', 'a')
            if product_download_vector:
                product_download_vector = list(product_download_vector.absolute_links)[0]

            product = {'delivery time': product_delivery_time,
                       'version': product_version,
                       'status': product_status,
                       'download link pdf': product_download_pdf,
                       'download link jpeg': product_download_jpeg,
                       'download link vector': product_download_vector}
            component_products[product_name] = product

        return component_products

    def __scrape_areas_of_interest(self, activation_code, component_html):
        """
        Determines all areas of interest for the given component and scrapes its corresponding products
        """
        areas_of_interest = {}
        aoi_filter, _ = self.__search_by_waypoints(component_html, '.view-components-list-product')
        # if no separate areas of interest (AoIs) are defined, create default ALL AoI
        if aoi_filter is None:
            products_root = component_html.find('.view-content')[-1]
            products = self.__scrape_component_products(activation_code, products_root)
            areas_of_interest['ALL'] = products
        else:
            aoi_filter_links = aoi_filter.absolute_links
            for aoi_filter_link in aoi_filter_links:
                aoi_name = aoi_filter_link.split("/")[-1]
                try:
                    response = self.__send_request(aoi_filter_link)
                    products_root = response.html.find('.view-content')[-1]
                    products = self.__scrape_component_products(activation_code, products_root)
                    areas_of_interest[aoi_name] = products
                except:
                    areas_of_interest[aoi_name] = {}
        return areas_of_interest

    def scrape_activation_component(self, url):
        """Scrapes all information from a component page including meta information, areas of interest and
        corresponding lists of products
        """
        try:
            response = self.__send_request(url)
            response.raise_for_status()

            html = response.html
            meta_root = html.find('.view-content', first=True)
            activation_code = url.split('/')[-1]
            component = self.__scrape_meta_root_fields(activation_code, meta_root)
            areas_of_interest = self.__scrape_areas_of_interest(activation_code, html)
            component["areas of interest"] = areas_of_interest
            component["component url"] = url
            return component, ""
        except Exception as e:
            return {}, str(e)

    def scrape_activation_components(self, **kwargs):
        activation_components = {}
        if self.activations is None:
            success, error_msg = self.scrape_activations_list_url(**kwargs)
            if not success:
                print(error_msg)
                return None, error_msg

        num_components_found = len(self.activations)

        if self.progress_report:
            step_message = "Found {} components matching given filters".format(num_components_found)
            scraping_report = self.progress_report.create_subreport(step_message=step_message,
                                                                    steps=num_components_found)

        for idx, row in self.activations.iterrows():
            if self.progress_report:
                step_message = "Scraping component {} of {}".format(idx+1, num_components_found)
                comp_report = scraping_report.create_subreport(step_message=step_message)
            component_url = row['Component List Link']
            current_component, _ = self.scrape_activation_component(component_url)
            if current_component:
                component_id = current_component['component id']
                activation_components[component_id] = current_component

            if self.progress_report:
                step_message = "Scraped component {} of {}".format(idx+1, num_components_found)
                comp_report.step_message = step_message
                comp_report.complete()
            wait_time = randint(self.random_wait_time_min_sec, self.random_wait_time_max_sec)
            time.sleep(wait_time)

        if self.progress_report:
            scraping_report.step_message = "Scraping procedure finished."
            scraping_report.complete()

        return activation_components, ""
