from datetime import datetime
import json
import io
import zipfile
import re
import os


class ActivationProduct:

    def __init__(self, product_name, product_dict):
        self.name = product_name
        delivery_time_string = product_dict['delivery time'].replace("(UTC)", "").strip()
        self.timestamp = datetime.strptime(delivery_time_string, "%Y-%m-%d %H:%M:%S")  # 2021-02-09 11:57:50
        self.version = product_dict['version']
        self.status = product_dict['status']
        self.vector_source = product_dict['download link vector']
        self.pdf_source = product_dict['download link pdf']
        self.jpeg_source = product_dict['download link jpeg']
        self.json = {product_name: product_dict}


class AreaOfInterest:

    def __init__(self, aoi_name, aoi_info, sort_products_most_recent_first=True):
        self.name = aoi_name
        self.products = []
        for product_name, product_dict in aoi_info.items():
            try:
                product = ActivationProduct(product_name, product_dict)
                self.products.append(product)
            except Exception as e:
                print("Invalid product: {}".format(product_name))
                print("Creation of product failed with exception: {}. Skipping...".format(str(e)))
                continue
        self.__sorted = False
        if sort_products_most_recent_first:
            self.sort_products_by_timestamp(reverse=True, override_current_order=True)
            self.__sorted = True
        self.json = {aoi_name: aoi_info}

    def sort_products_by_timestamp(self, reverse=False, override_current_order=False):
        products_with_timestamps = []
        for product in self.products:
            timestamp = product.timestamp
            products_with_timestamps.append((timestamp, product))
        products_with_timestamps = sorted(products_with_timestamps, key=lambda t: t[0], reverse=reverse)
        sorted_products = [product for _, product in products_with_timestamps]
        if override_current_order:
            self.products = sorted_products
            self.__sorted = True
            return sorted_products
        else:
            return sorted_products

    def get_latest_vector_product(self):
        if not self.__sorted:
            sorted_products = self.sort_products_by_timestamp(reverse=True)
        else:
            sorted_products = self.products
        for product in sorted_products:
            if product.vector_source:
                return product
        return None

    def get_latest_pdf_product(self):
        if not self.__sorted:
            sorted_products = self.sort_products_by_timestamp(reverse=True)
        else:
            sorted_products = self.products
        for product in sorted_products:
            if product.pdf_source:
                return product
        return None

    def get_latest_jpeg_product(self):
        if not self.__sorted:
            sorted_products = self.sort_products_by_timestamp(reverse=True)
        else:
            sorted_products = self.products
        for product in sorted_products:
            if product.jpeg_source:
                return product
        return None


class ActivationComponent:

    def __init__(self, component_dict):
        self.id = component_dict['component id']
        self.meta = component_dict['meta']
        areas_of_interest_dict = component_dict['areas of interest']
        self.areas_of_interest = {aoi_name: AreaOfInterest(aoi_name, aoi_info)
                                  for aoi_name, aoi_info in areas_of_interest_dict.items()}
        self.json = {self.id: component_dict}


class FloodUpdateManager:
    """
    Utility class to manage updates for the geoserver
    """

    def __init__(self, geoserver, workspace, scraper, current_components=None, latest_components=None):
        self.geoserver = geoserver
        self.workspace = workspace
        self.scraper = scraper
        self.current_components_index = current_components
        self.latest_components_index = latest_components
        self.update_index = None

    def index_latest_components(self, **kwargs):
        if self.scraper.activations is None:
            if kwargs is None:
                suc, err = self.scraper.scrape_activations_list_url()
            else:
                suc, err = self.scraper.scrape_activations_list_url(**kwargs)

            if not suc:
                return False, err
        else:
            latest_components_dict, err = self.scraper.scrape_activation_components()
            if not latest_components_dict:
                return False, err
            else:
                self.latest_components_index = {ActivationComponent(component).id: ActivationComponent(component)
                                                for component in latest_components_dict.values()}
                return True, ""

    def index_current_components(self):
        data_stores = self.geoserver.get_datastores(workspace=self.workspace)
        if type(data_stores) != dict:
            return False, data_stores
        elif not data_stores['dataStores']:
            self.current_components_index = {}
            return True, ""
        else:
            for data_store in data_stores['dataStores']['dataStore']:
                data_store_name = data_store['name']
                data_store = self.geoserver.get_datastore(store_name=data_store_name,
                                                          workspace=self.workspace)['dataStore']
                if type(data_store) != dict:
                    return False, data_store
                else:
                    try:
                        description = json.loads(data_store['description'])
                        component_dict = list(description.values())[0]
                        if self.current_components_index is None:
                            self.current_components_index = {}
                        if not self.current_components_index.get(component_dict["component id"], None):
                            component = ActivationComponent(component_dict)
                            component_id = component.id
                            self.current_components_index[component_id] = component
                    except json.decoder.JSONDecodeError:
                        print("Error: datastore {} without description.".format(data_store_name))
                        print("Please copy over description manually from other datastore of same component or delete" +
                              " all datastores from the same component.")
                        continue
                    except Exception as e:
                        print("Exception during indexation of current components: " + str(e))
            return True, ""

    def build_update_index(self):
        if not self.current_components_index:
            suc, err = self.index_current_components()
            if not suc:
                return suc, err
        if not self.latest_components_index:
            suc, err = self.index_latest_components()
            if not suc:
                return suc, err

        if self.update_index is None:
            self.update_index = {}

        for comp_id, comp in self.latest_components_index.items():
            comp_old = self.current_components_index.get(comp_id, None)
            update_meta = {}
            # if completely new component, add latest products for all areas of interest to update index
            if not comp_old:
                update_meta['new component'] = comp
                update_meta['previous component'] = None
                update_meta['deprecated areas of interest'] = {}
                updates = {}
                for aoi in comp.areas_of_interest.values():
                    latest_vector_product = aoi.get_latest_vector_product()
                    if latest_vector_product is not None:
                        updates[aoi.name] = latest_vector_product
                if updates:
                    update_meta['updates'] = updates
                    self.update_index[comp_id] = update_meta
                continue

            else:
                update_meta['new component'] = comp
                update_meta['previous component'] = comp_old
                update_meta['deprecated areas of interest'] = {}
                updates = {}
                for aoi in comp.areas_of_interest.values():
                    latest_vector_product = aoi.get_latest_vector_product()
                    if latest_vector_product is None:
                        continue

                    aoi_old = comp_old.areas_of_interest.get(aoi.name, None)
                    if aoi_old is None:
                        updates[aoi.name] = latest_vector_product
                        continue

                    prev_vector_product = aoi_old.get_latest_vector_product()
                    if latest_vector_product.name != prev_vector_product.name:
                        updates[aoi.name] = latest_vector_product
                        update_meta['deprecated areas of interest'][aoi_old.name] = prev_vector_product

                if updates:
                    update_meta['updates'] = updates
                    self.update_index[comp_id] = update_meta
                continue
        return True, ""

    def fetch_aoi_update_for_component(self, component_id, aoi_name, retries=2):
        if self.update_index is None:
            suc, err = self.build_update_index()
            if not suc:
                return None, err

        update_meta = self.update_index.get(component_id, None)
        if not update_meta:
            return None, "Component is already up to date"

        aoi_latest_product = update_meta['updates'].get(aoi_name, None)
        if not aoi_latest_product:
            return None, "AoI is already up to date"
        else:
            err = ""
            for retry in range(retries + 1):
                product_response = self.scraper.request_product_data(aoi_latest_product.vector_source)
                try:
                    product_response.raise_for_status()
                    product_bytes = io.BytesIO(product_response.content)
                    is_valid_zipfile = zipfile.is_zipfile(product_bytes)
                    if not is_valid_zipfile:
                        if retry == retries:
                            err = "Product Error: Invalid or broken zipfile"
                            return None, err
                        else:
                            continue
                    else:
                        return product_bytes, err
                except Exception as e:
                    if retry == retries:
                        err = str(e)
                        return None, err
                    else:
                        continue

    @classmethod
    def update_is_valid(cls, original_zipfile, keep_reg_exp, mandatory_file_ext=None):
        """

        :param original_zipfile: zip archive of downloaded product
        :param keep_reg_exp: regular expression defining which files to keep from original zip archive
        :param mandatory_file_ext: which file extensions should be present in the new archive after filtering. If set
            to None only empty filtered archives are considered invalid
        :return: True if filtered archive contains at least one file for each mandatory file extension and is not empty
            otherwise False
        """
        try:
            original_zipfile = zipfile.ZipFile(original_zipfile)
        except Exception as e:
            print(e)
            return False

        num_matches = 0
        if mandatory_file_ext is None:
            for zipinfo in original_zipfile.filelist:
                if re.fullmatch(keep_reg_exp, zipinfo.filename):
                    print("Found mandatory update file: ", zipinfo.filename)
                    num_matches += 1
            return num_matches > 0
        else:
            ext_found_dict = {}
            for file_ext in mandatory_file_ext:
                file_ext_re = file_ext.replace(".", "[.]") + "$"  # match dot literally and only at the end of string
                ext_found_dict[file_ext_re] = False

            for zipinfo in original_zipfile.filelist:
                if re.fullmatch(keep_reg_exp, zipinfo.filename):
                    num_matches += 1
                    print("Found mandatory update file: ", zipinfo.filename)
                    for file_ext_re in ext_found_dict:
                        match = re.search(file_ext_re, zipinfo.filename)
                        if match:
                            ext_found_dict[file_ext_re] = True
            all_ext_found = all(list(ext_found_dict.values()))
            return num_matches and all_ext_found

    @classmethod
    def filter_product_zipfile(cls, original_zipfile, keep_reg_exp, ignore_directory_structure=True):
        """
            Param: original_zipfile -> zipfile archive object to filter from or path to zipfile
            Param: keep_reg_exp -> regular expression describing which ZipInfos to keep from the original archive
            Param: ignore_directory_structure -> if True all files that match the regular expression are copied to the
                root directory of the resulting filtered zipfile
            Returns a new in-memory-zip file only containing items with names fully matching the regular expression
            given by keep_reg_exp
            """
        try:
            original_zipfile = zipfile.ZipFile(original_zipfile)
        except Exception as e:
            print(e)
            return None

        new_zip_buffer = io.BytesIO()
        with zipfile.ZipFile(new_zip_buffer, 'w') as new_archive:
            for zipinfo in original_zipfile.filelist:
                filename = zipinfo.filename
                basename = os.path.basename(zipinfo.filename)
                if re.fullmatch(keep_reg_exp, basename):
                    if not ignore_directory_structure:
                        new_archive.writestr(filename, original_zipfile.read(filename),
                                             compress_type=zipfile.ZIP_DEFLATED)
                    else:
                        new_archive.writestr(basename, original_zipfile.read(filename),
                                             compress_type=zipfile.ZIP_DEFLATED)
        return new_zip_buffer
