import requests
import re
import xml.etree.ElementTree as ElementTree
from shapely.geometry import Polygon
import numpy as np
from rasterio.io import MemoryFile
from rasterio.windows import Window
import geopandas as gpd


class CoordinateReferenceSystemAxis:
    def __init__(self, axis_label, unit_of_measurement=None, order_index=None, lower_bound=None, upper_bound=None,
                 resolution=None):
        self.axis_label = axis_label
        self.uom = unit_of_measurement
        self.order_index = order_index
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.resolution = resolution

    def value_within_axis_bounds(self, value):
        return self.lower_bound <= value <= self.upper_bound

    def subset_within_axis_bounds(self, subset):
        return self.value_within_axis_bounds(subset[0]) and self.value_within_axis_bounds(subset[1])


class GeoserverWebCoverageService:

    def __init__(self, geoserver, coverage_id, version="2.0.0"):
        self.geoserver = geoserver
        self.coverage_id = coverage_id
        self.version = version

        try:
            response = self.request_coverage_description()
            response.raise_for_status()

            self.xmlRoot = ElementTree.fromstring(response.content)
            self.xmlns = self.__get_xml_namespaces(response.text)
            self.srsName = self.__get_attribute_from_xml_node(".//gml:Envelope", "srsName")
            self.epsgCode = int(self.srsName.split("/")[-1])

            self.wkt, err = self.get_wkt_for_epsg_code(self.epsgCode)

            axis_labels = self.__get_attribute_from_xml_node(".//gml:Envelope", "axisLabels").split(" ")
            # units of measurement
            uom_labels = self.__get_attribute_from_xml_node(".//gml:Envelope", "uomLabels").split(" ")

            self.axis = [CoordinateReferenceSystemAxis(axis_label, unit_of_measurement=axis_uom)
                         for (axis_label, axis_uom) in zip(axis_labels, uom_labels)]

            default_axis_order = list(range(len(axis_labels)))
            default_order_msg = "Defaulting to axis order: {}. Please change manually if necessary!"
            if self.wkt is None:
                axis_order = default_axis_order
                print("WARNING: Request for WKT of EPSG:{} failed.({}).".format(self.epsgCode, err))
                print(default_order_msg.format(axis_order))
            else:
                axis_order = self.get_axis_order_from_wkt(self.wkt)
                if axis_order is None:
                    print("WARNING: Axis labels in WKT do not match axis labels from coverage description.")
                    axis_order = default_axis_order
                    print(default_order_msg.format(axis_order))

            self.set_axis_order(axis_order)
            self.sort_axis_by_order_index()
            axis_bounds = self.__get_axis_bounds_from_xml()
            self.set_axis_bounds(axis_bounds)
            axis_resolutions = self.__get_axis_resolutions_from_xml()
            self.set_axis_resolutions(axis_resolutions)

        except requests.exceptions.HTTPError as e:
            print("Initialization of WCS not possible. Error: " + str(e))
            raise e
        except Exception as e:
            raise e

    @classmethod
    def __get_xml_namespaces(cls, xml_string):
        namespaces = re.findall(r"\sxmlns:.*?\s", xml_string)
        xml_namespaces = {}
        for ns in namespaces:
            ns = ns.strip()
            ns_uri = ns.split("=")[1].strip('"')
            ns_name = ns.split("=")[0].split(":")[1]
            xml_namespaces[ns_name] = ns_uri
        return xml_namespaces

    def __get_attribute_from_xml_node(self, node_name, attr_name):
        node = self.xmlRoot.find(node_name, self.xmlns)
        attribute = node.attrib[attr_name]
        return attribute

    def __get_axis_bounds_from_xml(self):
        lower_bounds = self.xmlRoot.find(".//gml:lowerCorner", self.xmlns).text.split(" ")
        lower_bounds = [float(lower_bound) for lower_bound in lower_bounds]
        upper_bounds = self.xmlRoot.find(".//gml:upperCorner", self.xmlns).text.split(" ")
        upper_bounds = [float(upper_bound) for upper_bound in upper_bounds]
        bounds = list(zip(lower_bounds, upper_bounds))
        return bounds

    def __get_axis_resolutions_from_xml(self):
        offset_vector_tags = self.xmlRoot.findall(".//gml:offsetVector", self.xmlns)
        offset_vectors = []
        for offsetVectorTag in offset_vector_tags:
            offset_vector = list(map(lambda x: abs(float(x)), offsetVectorTag.text.split(" ")))
            offset_vectors.append(offset_vector)
        axis_resolutions = [sum(components) for components in zip(*offset_vectors)]
        return axis_resolutions

    def sort_axis_by_order_index(self):
        axis_sorted = sorted(self.axis, key=lambda ax: ax.order_index)
        self.axis = axis_sorted

    @classmethod
    def get_wkt_for_epsg_code(cls, epsg_code):
        service_url_template = r"https://apps.epsg.org/api/v1/CoordRefSystem/{}/export/?format=WKT"
        service_url = service_url_template.format(epsg_code)
        try:
            response = requests.get(service_url)
            response.raise_for_status()
            return response.text, ""
        except requests.exceptions.HTTPError:
            return None, "EPSG API request failed with status code: {}".format(response.status_code)

    def get_axis_order_from_wkt(self, wkt, wcs_axis_to_wkt_axis_map=None):
        axs = re.findall(r"AXIS\[.*?\]", wkt)
        axs = [re.search(r"\(\w+\)", ax).group(0)[1:-1] for ax in axs]
        order = []
        for ax in self.axis:
            try:
                if wcs_axis_to_wkt_axis_map is not None:
                    try:
                        wkt_axis_label = wcs_axis_to_wkt_axis_map[ax.axis_label]
                    except KeyError:
                        wkt_axis_label = ax.axis_label
                    axis_label = wkt_axis_label
                else:
                    axis_label = ax.axis_label

                idx = axs.index(axis_label)
                order.append(idx)
            except ValueError:
                print("WARNING: Mismatch between WKT axis names and WCS axis names.")
                return None
        return order

    def epsg_crs_is_xy(self, epsgCode, wkt=None):
        if wkt is None:
            wkt, msg = self.get_wkt_for_epsg_code(epsgCode)
            if wkt is None:
                print("WARNING: Defaulting to False. WKT request failed with error message: " + msg)
                return False
        axs = re.findall(r"AXIS\[.*?\]", wkt)
        axs = [re.search(r"\(\w+\)", ax).group(0)[1:-1] for ax in axs]
        return axs == ["X", "Y"] or axs == ["Lon", "Lat"]

    def set_axis_order(self, axis_order):
        assert len(self.axis) == len(axis_order), "Length of axis_order must match axis count"
        for ax, order_idx in zip(self.axis, axis_order):
            ax.order_index = order_idx

    def set_axis_bounds(self, axis_bounds):
        assert len(self.axis) == len(axis_bounds), "Length of axis_bounds must match axis count"
        for ax, (lower, upper) in zip(self.axis, axis_bounds):
            ax.lower_bound = lower
            ax.upper_bound = upper

    def set_axis_resolutions(self, axis_resolutions):
        assert len(self.axis) == len(axis_resolutions), "Length of axis_resolutions must match axis count"
        for ax, resolution in zip(self.axis, axis_resolutions):
            ax.resolution = resolution

    def create_wcs_request_url(self, request_type, coverage_identifier, output_format=None, subsets=None,
                               subsets_crs=None, scale_factor=None, range_subset=None, output_crs=None):
        base_url = "{}/wcs?service=WCS&version={}&request={}&coverageId={}".format(
            self.geoserver.service_url,
            self.version,
            request_type,
            coverage_identifier
        )

        request_url = base_url

        if output_format:
            request_url += "&format={}".format(output_format)

        if subsets:
            for ax in self.axis:
                axis_label = ax.axis_label
                subset_bounds = subsets.get(axis_label, None)
                if subset_bounds is not None:
                    lower = subset_bounds[0]
                    upper = subset_bounds[1]
                    request_url += "&subset={}({},{})".format(axis_label, lower, upper)

        if subsets and subsets_crs:
            request_url += "&subsettingCRS={}".format(subsets_crs)

        if scale_factor:
            request_url += "&scalefactor={}".format(scale_factor)

        if range_subset:
            request_url += "&rangesubset={}".format(range_subset)

        if output_crs:
            request_url += "&outputCRS={}".format(output_crs)

        return request_url

    def subsets_are_valid(self, subsets):
        valid = True
        for ax in self.axis:
            axis_label = ax.axis_label
            subset = subsets.get(axis_label, None)
            if subset:
                subset_within_axis_bounds = ax.subset_within_axis_bounds(subset)
                if not subset_within_axis_bounds:
                    valid = False
        return valid

    def request_coverage_description(self):
        service_url = self.create_wcs_request_url("DescribeCoverage", self.coverage_id)
        response = requests.get(service_url, auth=(self.geoserver.username, self.geoserver.password))
        response.raise_for_status()
        return response

    def request_coverage_data(self, output_format=None, subsets=None, subsets_crs=None, scale_factor=None,
                              range_subset=None, output_crs=None):
        service_url = self.create_wcs_request_url("GetCoverage", self.coverage_id, output_format=output_format,
                                                  subsets=subsets, subsets_crs=subsets_crs, scale_factor=scale_factor,
                                                  range_subset=range_subset, output_crs=output_crs)
        response = requests.get(service_url, auth=(self.geoserver.username, self.geoserver.password))
        response.raise_for_status()
        return response


class GeoserverWebFeatureService:

    def __init__(self, geoserver, feature_id, version="2.0.0"):
        self.geoserver = geoserver
        self.feature_id = feature_id
        self.version = version

    def create_wfs_request_url(self, request_type, feature_identifier, output_format=None, format_options=None):
        base_url = "{}/wfs?service=WFS&version={}&request={}&typeName={}".format(
            self.geoserver.service_url,
            self.version,
            request_type,
            feature_identifier
        )

        request_url = base_url

        if output_format:
            request_url += "&outputFormat={}".format(output_format)

        if format_options:
            format_options_string = ";".join(["{}:{}".format(k, v) for k, v in format_options.items()])
            request_url += "&format_options={}".format(format_options_string)

        return request_url

    def request_feature_description(self):
        service_url = self.create_wfs_request_url("DescribeFeatureType", self.feature_id)
        response = requests.get(service_url, auth=(self.geoserver.username, self.geoserver.password))
        response.raise_for_status()
        return response

    def request_feature_data(self, output_format=None, format_options=None):
        service_url = self.create_wfs_request_url("GetFeature", self.feature_id, output_format=output_format,
                                                  format_options=format_options)
        response = requests.get(service_url, auth=(self.geoserver.username, self.geoserver.password))
        response.raise_for_status()
        return response

    def request_feature_geodataframe(self, split_multipolygons=False, output_format="shape-zip"):
        service_url = self.create_wfs_request_url("GetFeature", self.feature_id, output_format=output_format)
        dataframe = gpd.read_file(service_url)
        if split_multipolygons:
            # remove Multipolygons from dataframe by expanding them to multiple single polygon rows
            dataframe = dataframe.explode().reset_index()
            drop_columns = [col for col in dataframe.columns if "level_" in col]
            dataframe = dataframe.drop(drop_columns, axis=1)
        return dataframe


class DigitalElevationModelWCS(GeoserverWebCoverageService):

    def __init__(self, *args, **kwargs):
        super(DigitalElevationModelWCS, self).__init__(*args, **kwargs)

    @classmethod
    def create_raster_dataset_from_bytes(cls, data_bytes):
        """Returns a rasterio dataset given raster data in bytes."""
        with MemoryFile(data_bytes) as memory_file:
            return memory_file.open()

    @classmethod
    def densify_polygon_by_interval(cls, polygon, spacing=1, is_percent=False):
        """Add points, at a fixed spacing, to a polygon.

        Parameters
        ----------
        polygon : shapely polygon to densify
        spacing : number
            Spacing between the points to be added to the line.
        is_percent : boolean
            Express the densification as a percent of the total length.

        Source:
        https://stackoverflow.com/questions/64995977/generating-equidistance-points-along-the-boundary-of-a-polygon-but-cw-ccw

        """
        assert spacing > 0, "Spacing must be > 0"

        a = np.array(polygon.exterior.coords)

        N = len(a) - 1  # segments
        dxdy = a[1:, :] - a[:-1, :]  # coordinate differences
        # fancy way to compute row-wise norm of an array
        leng = np.sqrt(np.einsum('ij,ij->i', dxdy, dxdy))  # segment lengths
        if is_percent:  # as percentage
            spacing = abs(spacing)
            spacing = min(spacing / 100, 1.)
            steps = (sum(leng) * spacing) / leng  # step distance
        else:
            steps = leng / spacing  # step distance
        deltas = dxdy / (steps.reshape(-1, 1))  # coordinate steps
        pnts = np.empty((N,), dtype='O')  # construct an `O` array
        for i in range(N):  # cycle through the segments and make
            num = np.arange(steps[i])  # the new points
            pnts[i] = np.array((num, num)).T * deltas[i] + a[i]
        a0 = a[-1].reshape(1, -1)  # add the final point and concatenate
        a = np.concatenate((*pnts, a0), axis=0)

        densified_polygon = Polygon(a)
        return densified_polygon

    @classmethod
    def sample_raster_at_polygon_perimeter(cls, polygon, raster_dataset):
        sample_positions = list(polygon.exterior.coords)
        samples = np.array(list(raster_dataset.sample(sample_positions))).flatten()
        return samples

    @classmethod
    def __rolling_window(cls, a, window_size):
        # source: http://www.mail-archive.com/numpy-discussion@scipy.org/msg29450.html
        shape = a.shape[:-1] + (a.shape[-1] - window_size + 1, window_size)
        strides = a.strides + (a.strides[-1],)
        return np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)

    @classmethod
    def __circular_rolling_window(cls, a, window_size):
        # source: https://stackoverflow.com/questions/35221645/rolling-window-on-a-circular-array
        pseudocircular = np.pad(a, pad_width=(0, window_size - 1), mode='wrap')
        return cls.__rolling_window(pseudocircular, window_size)

    @classmethod
    def __add_elevation_to_polygon_dem_smoothing(cls, polygon, raster_dataset, smoothing_kernel=None):
        poly_vertices = list(polygon.exterior.coords)
        if polygon.has_z:
            poly_vertices = [(x, y) for (x, y, z) in poly_vertices]

        if smoothing_kernel is None:
            elevation_levels = cls.sample_raster_at_polygon_perimeter(polygon, raster_dataset)
        else:
            elevation_levels = []
            for x, y in poly_vertices:
                row, col = raster_dataset.index(x, y)
                window_height, window_width = smoothing_kernel.shape

                left_bound = col - np.floor(window_width / 2)
                right_bound = col + np.ceil(window_width / 2)
                upper_bound = row - np.floor(window_height / 2)
                lower_bound = row + np.ceil(window_height / 2)

                if left_bound < 0:
                    smoothing_kernel = smoothing_kernel[-left_bound:]
                    col_off = 0
                    window_width += left_bound
                else:
                    col_off = left_bound

                if upper_bound < 0:
                    smoothing_kernel = smoothing_kernel[-upper_bound:]
                    row_off = 0
                    window_height += upper_bound
                else:
                    row_off = upper_bound

                if right_bound > raster_dataset.width:
                    delta = raster_dataset.width - right_bound
                    smoothing_kernel = smoothing_kernel[:delta]
                    window_width += delta

                if lower_bound > raster_dataset.height:
                    delta = raster_dataset.height - lower_bound
                    smoothing_kernel = smoothing_kernel[:delta]
                    window_height += delta

                window = Window(col_off, row_off, window_width, window_height)
                elevation_patch = raster_dataset.read(1, window=window)
                elevation_level = np.sum(smoothing_kernel * elevation_patch)
                elevation_levels.append(elevation_level)

        poly_with_elevation = [(*coords, elevation) for coords, elevation in zip(poly_vertices, elevation_levels)]
        return Polygon(poly_with_elevation)

    @classmethod
    def __add_elevation_to_polygon_perimeter_smoothing(cls, polygon, raster_dataset, smoothing_kernel=None):
        elevation_levels = cls.sample_raster_at_polygon_perimeter(polygon, raster_dataset)
        number_of_vertices = elevation_levels.shape[0]
        if smoothing_kernel is not None:
            window_size = len(smoothing_kernel)
            if window_size > number_of_vertices:
                print("WARNING: window size > number of vertices. Window center-cropped to number of vertices.")
                delta_length = window_size - number_of_vertices
                smoothing_window = smoothing_kernel[np.floor(delta_length/2):-np.ceil(delta_length/2)]
                window_size = number_of_vertices

            rolling_window = cls.__circular_rolling_window(elevation_levels, window_size).reshape(-1, window_size)
            # shift such that first value in center of window
            rolling_window = np.roll(rolling_window, window_size//2, axis=0)
            elevation_levels = np.sum(rolling_window * smoothing_window, axis=1)

        polygon_xy = list(polygon.exterior.coords)
        if polygon.has_z:
            polygon_xy = [(x, y) for (x, y, z) in polygon_xy]
        polygon_xyz = [(*xy, z) for xy, z in zip(polygon_xy, elevation_levels)]
        return Polygon(polygon_xyz)

    @classmethod
    def add_elevation_to_polygon(cls, polygon, raster_dataset, smoothing_kernel=None):
        if smoothing_kernel is None:
            poly_z = cls.__add_elevation_to_polygon_perimeter_smoothing(polygon, raster_dataset)
        else:
            if isinstance(smoothing_kernel, list):
                smoothing_kernel = np.array(smoothing_kernel)

            if len(smoothing_kernel.shape) == 1:
                poly_z = cls.__add_elevation_to_polygon_perimeter_smoothing(polygon, raster_dataset,
                                                                            smoothing_kernel=smoothing_kernel)
            elif len(smoothing_kernel) == 2:
                poly_z = cls.__add_elevation_to_polygon_dem_smoothing(polygon, raster_dataset,
                                                                      smoothing_kernel=smoothing_kernel)
            else:
                assert 0 < len(smoothing_kernel) < 3, "Only kernels of dimension 1 or 2 are supported"
        return poly_z

    def add_elevation_to_geodataframe(self, geodataframe, raster_dataset, smoothing_kernel=None, densify=True):
        geodataframe_copy = geodataframe.copy()  # iterrows is not inplace
        for idx, row in geodataframe.iterrows():
            poly = row["geometry"]
            wcs_resolution = self.axis[0].resolution
            if densify:
                poly = self.densify_polygon_by_interval(poly, spacing=wcs_resolution)
            poly_z = self.add_elevation_to_polygon(poly, raster_dataset, smoothing_kernel=smoothing_kernel)
            geodataframe_copy.loc[idx, "geometry"] = poly_z
        return geodataframe_copy


