from typing import List, Optional
from deprogressapi import ProgressReport


class Stage:

    def __init__(self, name, function, description="", evaluator_fn=None, preprocessor_fn=None):
        """

        :param name: [str] human readable name for the stage
        :param function: [callable] functionality of the stage. will be executed when execute() is called
        :param description: [str] optional description containing additional info about the stage for users
        :param evaluator_fn: [callable] function that will be called on the output of the stage to check whether the
            output is considered to be as expected or not. must evaluate to true or false
        :param preprocessor_fn: [callable] function that allows to adapt output of previous stages to desired input
            format
        :param progress_report: [ProgressReport] progress report object to pass messages to during execution of stage
        """
        self.name = name
        self.function = function
        self.description = description

        if evaluator_fn is None:
            self.evaluator_fn = lambda x: True
        else:
            self.evaluator_fn = evaluator_fn

        self.preprocessor_fn = preprocessor_fn

        self.__output = None

    def execute(self, stage_input, progress_report=None):
        try:
            if not self.preprocessor_fn:
                self.__output = self.function(stage_input, progress_report=progress_report)
            else:
                stage_input = self.preprocessor_fn(stage_input)
                stage_input["progress_report"] = progress_report
                self.__output = self.function(**stage_input)
        except TypeError as e:
            error_message = "error: TypeError occured: \n {}".format(str(e))
            hint_message = \
                "Probably the provided stage function is missing the progress_report keyword even though" + \
                "a progress report object was passed to execute()"
            step_message = error_message + "\n" + hint_message
            if progress_report:
                sub_report = progress_report.create_subreport(step_message=step_message)
                sub_report.complete()
        except Exception as e:
            step_message = f'error: {str(e)}'
            if progress_report is not None:
                sub_report = progress_report.create_subreport(step_message=step_message)
                sub_report.complete()
            else:
                print(step_message)

    def evaluate_execution_status(self):
        try:
            execution_status = self.evaluator_fn(self.__output)
        except Exception as e:
            print("Evaluation of stage {} failed with Exception: {}".format(self.name, str(e)))
            execution_status = False
        return execution_status

    def get_output(self):
        return self.__output


class LinearPipeline:

    def __init__(self, name, stages: List[Stage], progress_report=None):
        self.name = name
        self.stages = stages
        self.__output = None
        self.__executed = False
        self.__initialized = False
        self.__progress_report = progress_report

    def set_progress_report(self, progress_report: ProgressReport):
        self.__progress_report = progress_report

    def initialize(self, first_stage_input):
        first_stage = self.stages[0]
        verbose = isinstance(self.__progress_report, ProgressReport)
        sub_report: Optional[ProgressReport] = None

        if verbose:
            sub_report = self.__progress_report.create_subreport(step_message="Initializing pipeline...")
            first_stage.execute(first_stage_input, progress_report=sub_report)
        else:
            first_stage.execute(first_stage_input)

        execution_status = first_stage.evaluate_execution_status()
        if execution_status:
            self.__initialized = True
            if verbose:
                step_message = "Initialization of pipeline: >{}< successful. Stage >{}< [1 of {}] executed."
                sub_report.step_message = step_message.format(self.name, first_stage.name, len(self.stages))
                sub_report.complete()
        else:
            if verbose:
                sub_report.step_message = "error: Initialization of pipeline: >{}< failed".format(self.name)
                sub_report.complete()

    def run(self):
        verbose = type(self.__progress_report) == ProgressReport
        if not self.__initialized:
            return self.__output, False

        if verbose:
            step_message = "Executing pipeline: >{}<".format(self.name)
            sub_report = self.__progress_report.create_subreport(step_message=step_message)

        if not self.stages:
            if verbose:
                step_message = "Pipeline is empty."
                sub_report.step_message = step_message
                sub_report.complete()
            self.__executed = True
        else:
            num_stages = len(self.stages)
            for stage_idx, stage in enumerate(self.stages):
                step_message = ("Executing stage: >{}< [{} of {}]...\n" +
                                "Stage description:\n{}").format(stage.name, stage_idx+1, num_stages,
                                                                 stage.description)
                if stage_idx:
                    if verbose:
                        sub_report = self.__progress_report.create_subreport(step_message=step_message)
                        stage.execute(self.__output, progress_report=sub_report)
                    else:
                        stage.execute(self.__output)

                execution_status = stage.evaluate_execution_status()
                self.__executed = execution_status
                self.__output = stage.get_output()
                if execution_status:
                    if verbose:
                        sub_report.step_message = "Execution of stage: >{}< successful".format(stage.name, stage_idx)
                        sub_report.complete()
                else:
                    if verbose:
                        sub_report.step_message = "error: Execution of stage: >{}< failed. Pipeline interrupted.".format(
                            stage.name, stage_idx)
                        sub_report.complete()
                    return self.__output, execution_status
            return self.__output, execution_status

    def initialize_and_run(self, first_stage_input):
        self.initialize(first_stage_input)
        output, status = self.run()
        return output, status

    def get_output(self):
        return self.__output

    def evaluate_execution_status(self):
        return self.__executed









