import json
from typing import Optional

import pyppeteer
import shapely
import shapely.geometry
from deprogressapi import ProgressReport
from requests_html import AsyncHTMLSession
from datetime import date, timedelta
from .digitalearth_geoserver import FloodGeoserver
from .digitalearth_scrapers import CopernicusScraper
from .digitalearth_update_managers import FloodUpdateManager
from .pipeline_utils import LinearPipeline, Stage
from .geoserver_web_service_utils import DigitalElevationModelWCS, GeoserverWebFeatureService


# GLOBALS
# Regular expression for which files to keep from Vector-ZIP packages
KEEP_PATTERN_RE = r'^(EMSR|EMNR)(-*)[0-9][0-9][0-9](.*?)(observedEvent|observed_event)(.*?)\.(shp|shx|dbf|prj)'
MANDATORY_FILE_EXT = KEEP_PATTERN_RE.split(".")[-1][1:-1].split("|")
# <,>, :, /, \, |, ?, * are illegal under windows
UNDERSCORE_CHARS = ["<", ">", ":", "/", "\\", "|", "?", "*", ",", "]", "#"]
REMOVE_CHARS = ["(", ")", "[", " "]


# HELPER CLASSES AND FUNCTIONS

# helper class for patching requests-html problems when not executed from main thread, taken from
# https://github.com/psf/requests-html/issues/293
class AsyncHTMLSessionFixed(AsyncHTMLSession):
    def __init__(self, **kwargs):
        super(AsyncHTMLSessionFixed, self).__init__(**kwargs)
        self.__browser_args = kwargs.get("browser_args", ["--no-sandbox"])

    @property
    async def browser(self):
        if not hasattr(self, "_browser"):
            self._browser = await pyppeteer.launch(ignoreHTTPSErrors=not self.verify, headless=True,
                                                   handleSIGINT=False, handleSIGTERM=False, handleSIGHUP=False,
                                                   args=self.__browser_args)
        return self._browser


# PIPELINE ROUTINE DEFINITIONS

def init_scraper_routine(config, progress_report=None):
    sub_report: Optional[ProgressReport] = None

    if progress_report:
        sub_report = progress_report.create_subreport(step_message="Attempting initialization of scraper.")

    base_url = config.get("base_url", None)
    activations_list_url = config.get("activations_list_url", None)
    min_recency_in_days = config.get("min_recency_in_days", None)
    activation_filters = config.get("activation_filters", {})

    if min_recency_in_days and min_recency_in_days > 0 and not activation_filters.get("event_time_utc_start", None):
        event_time_utc_start = date.today() + timedelta(days=-min_recency_in_days)
        activation_filters["event_time_utc_start"] = str(event_time_utc_start)

    config["activations_found"] = False
    if base_url and activations_list_url:
        session = AsyncHTMLSessionFixed()
        scraper = CopernicusScraper(base_url, activations_list_url, session=session)
        suc, err = scraper.scrape_activations_list_url(**activation_filters)
        if not suc:
            print("Activations scraping error: ", err)
        activations_found = scraper.activations is not None
        if activations_found:
            config["activations_found"] = True
            if sub_report:
                sub_report.step_message = "Scraper initialized."
                sub_report.complete()
        return config, scraper

    if sub_report:
        sub_report.step_message = \
            "Scraper couldn't be initialized. Please check configuration and network connectivity."
        sub_report.complete()
    return config, None


def init_update_manager_routine(config, scraper, progress_report=None):
    sub_report: Optional[ProgressReport] = None
    if progress_report:
        sub_report = progress_report.create_subreport(step_message="Attempting initialization of update manager.")

    # Init UpdateManager
    geoserver_url = config.get("geoserver_url", None)
    username = config.get("username", None)
    password = config.get("password", None)
    workspace = config.get("workspace", None)

    geoserver = FloodGeoserver(geoserver_url, username=username, password=password)
    updateManager = FloodUpdateManager(geoserver, workspace, scraper)

    # test the connection
    res = geoserver.get_workspace(workspace)
    connection_status = not(res is None)
    config["updateManager_initialized"] = connection_status

    if sub_report:
        if not connection_status:
            sub_report.step_message = \
                "Can't establish connection to geoserver. Please check configuration and network connectivity."
            sub_report.complete()
        else:
            sub_report.step_message = \
                "Update manager initialized."
            sub_report.complete()
    return config, updateManager


def build_current_components_index_routine(config, updateManager, progress_report=None):
    sub_report: Optional[ProgressReport] = None
    if progress_report:
        sub_report = progress_report.create_subreport(step_message="Attempting to build index of current components...")

    suc, err = updateManager.index_current_components()

    if sub_report:
        if not suc:
            sub_report.step_message = \
                "Can't build current components index: " + err
            sub_report.complete()
        else:
            sub_report.step_message = "Index of current components has been built."
            sub_report.complete()
    config["current_components_index_built"] = suc
    return config, updateManager


def build_latest_components_index_routine(config, updateManager, progress_report=None):
    sub_report: Optional[ProgressReport] = None
    if progress_report:
        sub_report = progress_report.create_subreport(step_message="Attempting to build index of latest components.")
        updateManager.scraper.progress_report = sub_report

    suc, err = updateManager.index_latest_components()
    if sub_report:
        if not suc:
            sub_report.step_message = \
                "Can't build latest components index. Please check configuration and network connectivity."
            sub_report.complete()
        else:
            sub_report.step_message = "Index of latest components has been built."
            sub_report.complete()
    config["latest_components_index_built"] = suc
    return config, updateManager


def build_update_index_routine(config, updateManager, progress_report=None):
    sub_report: Optional[ProgressReport] = None
    if progress_report:
        sub_report = progress_report.create_subreport(
            step_message="Attempting to build index of new and deprecated components.")

    suc, err = updateManager.build_update_index()

    if sub_report:
        if not suc:
            sub_report.step_message = \
                "Can't build update index. Please check configuration and network connectivity."
            sub_report.complete()
        else:
            sub_report.step_message = "Index of possible updates has been built."
            sub_report.complete()
    config["update_index_built"] = suc
    return config, updateManager


def push_updates_to_geoserver_routine(config, updateManager, progress_report=None):
    update_procedure_report: Optional[ProgressReport] = None
    if progress_report:
        update_procedure_report = progress_report.create_subreport(
            step_message="Attempting to upload components on geoserver.")

    deprecated_components_list = list(updateManager.update_index.keys())
    num_deprecated_components = len(deprecated_components_list)

    if update_procedure_report:
        if not num_deprecated_components:
            update_procedure_report.step_message = \
                "No updates or deprecated components found. Shut down update process."
            update_procedure_report.complete()
        else:
            step_message = "Found new/ deprecated components: " + \
                str(deprecated_components_list)
            update_procedure_report.step_message = step_message
            update_procedure_report.complete()

    if progress_report:
        update_procedure_report = progress_report.create_subreport(step_message="Running update procedure",
                                                                   steps=num_deprecated_components)

    component_description_updates = {}
    for comp_id, update_meta in updateManager.update_index.items():

        comp_report: Optional[ProgressReport] = None
        if update_procedure_report:
            step_message = "Running update sub-procedure for component: " + comp_id
            comp_report = update_procedure_report.create_subreport(step_message=step_message)
        updates = update_meta['updates']
        for aoi_name in updates:
            # don't update the AoI named "ALL" if its a component that is divided into more specific AoIs
            # "ALL" AoI is just a default for components that contain only a single AoI
            if len(updates) > 1 and aoi_name == "ALL":
                continue

            aoi_report: Optional[ProgressReport] = None
            if comp_report:
                step_message = "Attempt to create datastore for the following area of interest in component {}: {}".format(
                    comp_id, aoi_name)
                aoi_report = comp_report.create_subreport(step_message=step_message)

            product_in_memory_zip, err = updateManager.fetch_aoi_update_for_component(comp_id, aoi_name)

            if product_in_memory_zip is None:
                if aoi_report:
                    step_message = "Product update {} could not be downloaded: Error: {}".format(aoi_name, err)
                    aoi_report.step_message = step_message
                    aoi_report.complete()
                continue

            filtered_zip = updateManager.filter_product_zipfile(product_in_memory_zip, KEEP_PATTERN_RE)
            valid_update = updateManager.update_is_valid(filtered_zip, KEEP_PATTERN_RE,
                                                         mandatory_file_ext=MANDATORY_FILE_EXT)
            component_dict = update_meta['new component'].json
            if not component_description_updates.get(comp_id, None):
                component_description_updates[comp_id] = (component_dict, [])

            if valid_update:
                latest_product = updates[aoi_name]
                store_name = latest_product.name
                step_message = "Download of product data successful. Attempting to create datastore: " + store_name

                aoi_datastore_report: Optional[ProgressReport] = None
                if aoi_report:
                    aoi_datastore_report = aoi_report.create_subreport(step_message=step_message)
                workspace = config["workspace"]

                for uchar in UNDERSCORE_CHARS:
                    store_name = store_name.replace(uchar, "_")
                for rchar in REMOVE_CHARS:
                    store_name = store_name.replace(rchar, "")

                suc_msg = updateManager.geoserver.create_vector_datastore_from_in_memory_zip(filtered_zip, store_name,
                                                                                             workspace=workspace,
                                                                                             file_extension="shp")

                if aoi_datastore_report:
                    aoi_datastore_report.step_message = suc_msg
                    aoi_datastore_report.complete()

                suc = suc_msg == "The shapefile datastore created successfully!"

                if not suc:
                    # if datastore was not successfully created, delete the AoI from the component description
                    del component_description_updates[comp_id][0][comp_id]["areas of interest"][aoi_name]
                    if aoi_report:
                        aoi_report.step_message = "{}: creation of datastore not successful.".format(aoi_name)
                        aoi_report.complete()
                else:
                    component_description_updates[comp_id][1].append(store_name)
                    if aoi_report:
                        aoi_report.step_message = "{}: creation of datastore successful.".format(aoi_name)
                        aoi_report.complete()

            else:
                del component_description_updates[comp_id][0][comp_id]["areas of interest"][aoi_name]

                if aoi_report:
                    step_message = "{}: Update file does not contain the required files. Invalid update.".format(aoi_name)
                    aoi_report.step_message = step_message
                    aoi_report.complete()

        if comp_report:
            comp_report.step_message = "Component {} upload sub-procedure finished.".format(comp_id)
            comp_report.complete()

    if update_procedure_report:
        update_procedure_report.step_message = "Upload procedure finished."
        update_procedure_report.complete()

    config["component_description_updates"] = component_description_updates
    return config, updateManager


def update_datastore_descriptions_routine(config, updateManager, progress_report=None):
    description_updates = config.get("component_description_updates", {})
    workspace = config.get("workspace", None)
    updated_datastores = []

    update_report: Optional[ProgressReport] = None
    if progress_report:
        update_report = progress_report.create_subreport(
            step_message="Attempting to update description of all newly created datastores...")
    if not workspace:
        config["updated_datastores"] = updated_datastores
        if update_report:
            update_report.step_message = "No workspace given. Please check configuration."
            update_report.complete()
        return config, updateManager

    for component_dict, store_names in description_updates.values():
        description = json.dumps(component_dict)
        for store_name in store_names:
            store_report: Optional[ProgressReport] = None
            if update_report:
                store_report = update_report.create_subreport(
                    step_message="Updating datastore: {}...".format(store_name))

            status_code = updateManager.geoserver.update_datastore_description(workspace, store_name,
                                                                               description)
            if status_code in [200, 201]:
                if store_report:
                    store_report.step_message = "Datastore {} description updated.".format(store_name)
                    store_report.complete()
                updated_datastores.append((store_name, description))
            else:
                if store_report:
                    step_message = "WARNING: Datastore {} description update failed with status code: {}."
                    store_report.step_message = step_message.format(
                        store_name)
                    store_report.complete()

    config["updated_datastores"] = updated_datastores
    if update_report:
        update_report.step_message = "Description update procedure finished."
        update_report.complete()
    return config, updateManager


def add_elevations_routine(config, updateManager, progress_report=None):
    elevation_report: Optional[ProgressReport] = None
    if progress_report:
        elevation_report = progress_report.create_subreport(
            step_message="Attempting to add elevation levels to updated products.")

    workspace = config.get("workspace", None)
    updated_datastores = config.get("updated_datastores", None)

    if not updated_datastores:
        if elevation_report:
            elevation_report.step_message = "List of updated datastores is empty. Shutting down."
            elevation_report.complete()
        return config, updateManager

    dem = config.get("digital_elevation_model", None)
    smoothing_kernel = config.get("smoothing_kernel", None)
    elevation_suffix = config.get("elevation_suffix", "_3D")

    properly_configured = all([workspace, updated_datastores, dem])

    try:
        wcs_report: Optional[ProgressReport] = None
        if elevation_report:
            wcs_report = elevation_report.create_subreport(step_message="Attempting to initialize WCS...")
        wcs = DigitalElevationModelWCS(updateManager.geoserver, dem)
        if wcs.wkt is None:
            if wcs_report:
                step_message = "WARNING: EPSG API not responding, therefore axis ordering is unknown. This might lead " + \
                    "to issues with requesting subsets of the DEM if CRS axis order is not X (easting), Y (northing)."
                wcs_report.step_message = step_message
                wcs_report.complete()
        elif wcs.get_axis_order_from_wkt(wcs.wkt) is None:
            if wcs_report:
                step_message = "WARNING: Axis labels in WKT do not match axis labels from coverage description, " + \
                    "therefore axis ordering is unknown. This might lead " + \
                    "to issues with requesting subsets of the DEM if CRS axis order is not X (easting), Y (northing)."
                wcs_report.step_message = step_message
                wcs_report.complete()
        else:
            if wcs_report:
                wcs_report.step_message = "WCS initialized."
                wcs_report.complete()
    except Exception as e:
        if elevation_report:
            elevation_report.step_message = "Can't initialize WCS.\nException: {}".format(str(e)) + \
                "Adding elevation levels to updated products failed."
            elevation_report.complete()
        return config, updateManager

    wcs_resolution = wcs.axis[0].resolution
    if properly_configured:
        updated_datastores_copy = updated_datastores.copy()
        for product_name, description in updated_datastores:
            datastore = "{}:{}".format(workspace, product_name)

            product_report: Optional[ProgressReport] = None
            if elevation_report:
                step_message = "Starting elevation fetch sub-procedure for datastore: {}".format(datastore)
                product_report = elevation_report.create_subreport(step_message=step_message)

            wfs_report: Optional[ProgressReport] = None
            if product_report:
                wfs_report = product_report.create_subreport(step_message="Attempting to request data from WFS...")

            try:
                layer_name = updateManager.geoserver.get_featuretypes(workspace=workspace, store_name=product_name)[0]
            except IndexError:
                if wfs_report:
                    wfs_report.step_message = "No featuretypes available. Skipping datastore: {}".format(datastore)
                    wfs_report.complete()
                continue
            except Exception as e:
                print("Requesting featuretypes failed with exception: {}".format(str(e)))
                if wfs_report:
                    wfs_report.step_message = "Requesting featuretypes failed. Skipping datastore: {}".format(datastore)
                continue
            layer = "{}:{}".format(workspace, layer_name)
            wfs = GeoserverWebFeatureService(updateManager.geoserver, layer)
            try:
                event_df = wfs.request_feature_geodataframe(split_multipolygons=True)
            except Exception as e:
                if wfs_report:
                    wfs_report.step_message = "Attempt failed with error: {}".format(str(e))
                    wfs_report.complete()
                if product_report:
                    product_report.step_message = "Skipping datastore: {}".format(datastore)
                    product_report.complete()
                continue
            if wfs_report:
                wfs_report.step_message = "Received data from WFS."
                wfs_report.complete()

            enforce_xy_order = config.get("enforce_xy_order", False)
            if enforce_xy_order:
                epsgCode = int(event_df.crs["init"].split(":")[-1])
                wkt, err = wcs.get_wkt_for_epsg_code(epsgCode)
                if wkt is None:
                    if product_report:
                        product_report.step_message = \
                            "Skipping datastore {}. Can't determine axis order due to error: {}".format(datastore, err)
                        product_report.complete()
                    continue
                is_xy_order = wcs.epsg_crs_is_xy(epsgCode, wkt=wkt)
                if not is_xy_order:
                    event_df.geometry = event_df.geometry.map(lambda polygon: shapely.ops.transform(
                        lambda x, y: (y, x), polygon))
            event_df = event_df.to_crs(epsg=wcs.epsgCode)

            bbox = shapely.geometry.box(*event_df.total_bounds).buffer(wcs_resolution)
            x_min, y_min, x_max, y_max = bbox.bounds
            bounds = {
                "X": (x_min, x_max),
                "Lon": (x_min, x_max),
                "Y": (y_min, y_max),
                "Lat": (y_min, y_max)
            }

            axis_subsets = {wcs.axis[0].axis_label: bounds[wcs.axis[0].axis_label],
                            wcs.axis[1].axis_label: bounds[wcs.axis[1].axis_label]}

            subsets_are_valid = wcs.subsets_are_valid(axis_subsets)
            if not subsets_are_valid:
                if product_report:
                    product_report.step_message = "Event bounds are out-of-bounds for given DEM. " + \
                        "Please provide another DEM and/or check that the axis ordering is correct."
                    product_report.complete()
                continue

            response = wcs.request_coverage_data(subsets=axis_subsets)
            if response.status_code not in [200, 201]:
                if product_report:
                    product_report.step_message = "Can't fetch raster data, WCS not responding. Status Code: {}".format(
                        response.status_code
                    )
                    product_report.complete()
                continue
            raster_dataset = wcs.create_raster_dataset_from_bytes(response.content)

            event_df = wcs.add_elevation_to_geodataframe(event_df, raster_dataset, smoothing_kernel=smoothing_kernel)

            upload_report: Optional[ProgressReport] = None
            if progress_report:
                upload_report = progress_report.create_subreport(step_message="Attempting to upload modified event...")

            try:
                shp_name = "{}{}.shp".format(product_name, elevation_suffix)
                for uchar in UNDERSCORE_CHARS:
                    shp_name = shp_name.replace(uchar, "_")
                for rchar in REMOVE_CHARS:
                    shp_name = shp_name.replace(rchar, "")

                msg = updateManager.geoserver.create_shp_datastore_from_geodataframe(event_df, shp_name, workspace)

                if msg != "The shapefile datastore created successfully!":
                    if upload_report:
                        upload_report.step_message = "Upload failed due to error: {}".format(msg)
                        upload_report.complete()
                    if product_report:
                        product_report.step_message = "Skipping datastore: {}".format(datastore)
                        product_report.complete()
                    continue
                else:
                    try:
                        description_dict = json.loads(description)
                        [*description_dict.values()][0]["smoothing_kernel"] = str(smoothing_kernel)
                        description = json.dumps(description_dict)
                    except json.decoder.JSONDecodeError:
                        description += " used smoothing_kernel: {}".format(str(smoothing_kernel))
                    except Exception as e:
                        print("Adding smoothing_kernel to description failed with error: " + str(e))

                    new_ds_name = shp_name.replace(".shp", "")
                    status_code = updateManager.geoserver.update_datastore_description(workspace, new_ds_name,
                                                                                       description)
                    updated_datastores_copy.append((new_ds_name, description))
                    if status_code in [200, 201]:
                        if upload_report:
                            upload_report.step_message = "Upload successful."
                            upload_report.complete()
                        if product_report:
                            product_report.step_message = "Product procedure finished successfully."
                            product_report.complete()

                    else:
                        if upload_report:
                            step_message = "WARNING: datastore description could not be updated: Status Code{}".format(
                                status_code
                            )
                            upload_report.step_message = step_message
                            upload_report.complete()
                        if product_report:
                            product_report.step_message = "Product procedure finished incompletely.".format(datastore)
                            product_report.complete()
                        continue
            except Exception as e:
                if upload_report:
                    upload_report.step_message = "Upload failed due to error: {}".format(str(e))
                    upload_report.complete()
                if product_report:
                    product_report.step_message = "Skipping datastore: {}".format(datastore)
                    product_report.complete()
                continue

        if elevation_report:
            elevation_report.step_message = "Elevation procedure finished."
            elevation_report.complete()

    else:
        if elevation_report:
            elevation_report.step_message = "Request not properly configured. Please check that you provided " + \
                "correct values for 'workspace' and 'digital_elevation_model' in config dict"
            elevation_report.complete()

    config["updated_datastores"] = updated_datastores_copy
    return config, updateManager


def copy_datastore_descriptions_to_layers_routine(config, updateManager, progress_report=None):
    copy_report: Optional[ProgressReport] = None
    if progress_report:
        copy_report = progress_report.create_subreport(
            step_message="Attempting to copy datastore descriptions to their respective layers...")

    updated_datastores = config.get("updated_datastores", None)
    if not updated_datastores:
        if copy_report:
            copy_report.step_message = "No datastores were updated. Shutting down procedure"
            copy_report.complete()
        return None

    for ds_name, description in updated_datastores:
        ds_report: Optional[ProgressReport] = None
        if copy_report:
            ds_report = copy_report.create_subreport(step_message="Start copy procedure for datastore: {}".format(ds_name))

        try:
            layer_names = updateManager.geoserver.get_featuretypes(workspace=updateManager.workspace,
                                                                   store_name=ds_name)
        except Exception:
            if ds_report:
                ds_report.step_message = "Requesting featuretypes failed. Skipping datastore: {}".format(ds_name)
                ds_report.complete()
            continue

        if not layer_names:
            if ds_report:
                ds_report.step_message = "No featuretypes found for datastore: {}".format(ds_name)
                ds_report.complete()
            continue
        else:
            for layer_name in layer_names:
                layer_report: Optional[ProgressReport] = None
                if ds_report:
                    layer_report = ds_report.create_subreport(step_message="Update layer: {}".format(layer_name))
                status_code = updateManager.geoserver.update_featuretype_description(
                    workspace=updateManager.workspace,
                    store_name=ds_name,
                    featuretype_name=layer_name,
                    description=description)
                if layer_report:
                    if status_code in [200, 201]:
                        layer_report.step_message = "Layer {} updated".format(layer_name)
                    else:
                        layer_report.step_message = "Layer {} update failed with error code: {}".format(layer_name,
                                                                                                        status_code)
                    layer_report.complete()

        if ds_report:
            ds_report.step_message = "{}: copy procedure finished.".format(ds_name)
            ds_report.complete()

    if copy_report:
        copy_report.step_message = "Copy procedure finished."
        copy_report.complete()
    return None


# STAGE DEFINITIONS
description = "First initializes the scraper, then scrapes activations. Returns config and scraper object." + \
              "Stage is successful if activations are successfully scraped from website"
init_scraper_stage = Stage("Init Scraper",
                           function=init_scraper_routine,
                           evaluator_fn=lambda t: t[0]["activations_found"],
                           description=description)

description = "First initializes the updateManager, then tests connection to geoserver." + \
              "Returns config and updateManager object." + \
              "Stage is successful if updateManager gets created and connection to geoserver is established."
init_updateManager_stage = Stage("Init Update Manager",
                                 function=init_update_manager_routine,
                                 preprocessor_fn=lambda t: {"config": t[0], "scraper": t[1]},
                                 evaluator_fn=lambda t: t[0]["updateManager_initialized"],
                                 description=description)

description = "Attempts to build an index of all components & products currently uploaded on the geoserver" + \
              "Returns config and an update manager object. Stage is successful if index is successfully built."
index_current_components_stage = Stage("Index components & products on geoserver",
                                       function=build_current_components_index_routine,
                                       preprocessor_fn=lambda t: {"config": t[0], "updateManager": t[1]},
                                       evaluator_fn=lambda t: t[0]["current_components_index_built"],
                                       description=description)

description = "Attempts to build an index of all available components & products on the source website." + \
              "Returns config and an update manager object. Stage is successful if index is successfully built."
index_latest_components_stage = Stage("Index components & products on source website",
                                      function=build_latest_components_index_routine,
                                      preprocessor_fn=lambda t: {"config": t[0], "updateManager": t[1]},
                                      evaluator_fn=lambda t: t[0]["latest_components_index_built"],
                                      description=description)

description = "Attempts to build an index of all components that were previously unknown or deprecated." + \
              "Returns config and an update manager object. Stage is successful if index is successfully built."
index_updates_stage = Stage("Index components & products for which updates need to be downloaded.",
                            function=build_update_index_routine,
                            preprocessor_fn=lambda t: {"config": t[0], "updateManager": t[1]},
                            evaluator_fn=lambda t: t[0]["update_index_built"],
                            description=description)

description = "Downloads vector sources for new or deprecated components if available and searches them for" + \
    "required shapefiles. If they exist attempts to create a datastore for each product."
push_updates_stage = Stage("Fetch updates & create data stores", function=push_updates_to_geoserver_routine,
                           preprocessor_fn=lambda t: {"config": t[0], "updateManager": t[1]},
                           description=description)

description = "Updates the description of every newly created datastore."
update_descriptions_stage = Stage("Update datastore descriptions",
                                  function=update_datastore_descriptions_routine,
                                  preprocessor_fn=lambda t: {"config": t[0], "updateManager": t[1]},
                                  description=description)

description = "Attempts to add elevation levels from the digital elevation model to polygons for the given products" + \
    "A new datastore and a new layer are created for each product."
add_elevations_stage = Stage("Add elevation levels to updated event-polygons",
                             function=add_elevations_routine,
                             preprocessor_fn=lambda t: {"config": t[0], "updateManager": t[1]},
                             description=description)

description = "Attempts to copy the descriptions of all updated datastores to their respective layers."
add_layer_descriptions_stage = Stage("Update layer descriptions",
                                     function=copy_datastore_descriptions_to_layers_routine,
                                     preprocessor_fn=lambda t: {"config": t[0], "updateManager": t[1]},
                                     description=description)

# PIPELINE DEFINITION
stages = [init_scraper_stage, init_updateManager_stage, index_current_components_stage, index_latest_components_stage,
          index_updates_stage, push_updates_stage, update_descriptions_stage, add_elevations_stage,
          add_layer_descriptions_stage]
FLOOD_UPDATE_PIPELINE = LinearPipeline("digitalearth: flood workspace update pipeline", stages)

# USAGE EXAMPLE
if __name__ == "__main__":
    default_config = {
        "base_url": "https://emergency.copernicus.eu",
        "activations_list_url": "https://emergency.copernicus.eu/mapping/list-of-activations-rapid",
        "activation_filters": {"affected_countries": ["DE"]},
        "geoserver_url": "http://rz-vm154.gfz-potsdam.de:8080/geoserver",
        "username": "",
        "password": "",
        "workspace": "flood",
        "digital_elevation_model": "digitalearth:de_dem25m",
        "enforce_xy_order": True,
        "min_recency_in_days": 30
    }
    output, execution_status = FLOOD_UPDATE_PIPELINE.initialize_and_run(default_config)













